// GRANT module
//
//
module GRANT
#(
	parameter CHANNELS = 8, // total number of requistors
    parameter WIDTH = 32,   // the width of each requestor 's weight
    parameter WEIGHTLIMIT = 16
)
(
    reset,
    clk,
    scan_in0,
    scan_in1,
    scan_in2,
    scan_in3,
    scan_in4,
    scan_enable,
    test_mode,
    scan_out0,
    scan_out1,
    scan_out2,
    scan_out3,
    scan_out4,

    // input
    request,    // request input
    nextGrant,  // nextGrant from NGPRC
    weight,     // weight of current grant

    // output
    grant,      // grant output
);
input
    reset,      // system reset
    clk;        // system clock

input
    scan_in0,   // test scan mode data input
    scan_in1,   // test scan mode data input
    scan_in2,   // test scan mode data input
    scan_in3,   // test scan mode data input
    scan_in4,   // test scan mode data input
    scan_enable,// test scan mode enable
    test_mode;  // test mode
    
output
    scan_out0,  // test scan mode data output
    scan_out1,  // test scan mode data output
    scan_out2,  // test scan mode data output
    scan_out3,  // test scan mode data output
    scan_out4;  // test scan mode data output

// input
input [(CHANNELS-1):0] request;
input [(CHANNELS-1):0] nextGrant;
input [WIDTH-1:0] weight;

// output
output reg [(CHANNELS-1):0] grant;
    
// internal registers
reg [(WIDTH-1):0] s_counter;
reg [(CHANNELS-1):0] s_request;
reg [(WIDTH-1):0] s_weight;
// reg update;
//--- Internal Constants ---//
localparam SIZE = 4;

// STATES
reg [(SIZE-1):0] state;
localparam RESET = 4'b0001;         // 'b0001
localparam GRANT_PROCESS= 4'b0010;  // 'b0010
localparam COUNT = 4'b0100;         // 'b0100
localparam GETWEIGHT= 4'b1000;      // 'b1000

//--- Code Starts ---//
// registeri/delay request
always @(posedge clk, posedge reset) begin : requestDelay
    if (reset) begin
        s_request <= 0;
    end else begin
        s_request <= request;
    end
end

// always block for state transition
always @(posedge clk, posedge reset) begin : stateTransition
    // reset condition
    if (reset) begin
        state = 0;
    end else begin
        case (state)
            // grant process this output grant
            GRANT_PROCESS: begin
                // if there is request
                // go to COUNT state to count
                if (grant != 0)
                    state = GETWEIGHT;
                // just stay here and process next
                else
                    state = state;
            end
            GETWEIGHT: begin
                state = COUNT;
            end

            // count clock cycle according to weight
            COUNT: begin
                // if counter is up
                // move to grant next
                if (s_counter >= s_weight)
                    state = GRANT_PROCESS;
                // fairness limit set by user
                // default is 16
                else if (s_counter >= WEIGHTLIMIT)
                    state = GRANT_PROCESS;
                // else keep couting
                else
                    state = state;
            end

            // if statemachine never goes out of wack
            default: begin
                state = GRANT_PROCESS;
            end
        endcase
    end
end

// output logic
always @(posedge clk, posedge reset) begin : grantStateMachineOutputLogic
    if (reset == 1'b1) begin
        grant = 0;
        s_counter = 0;
        s_weight = 0;
    end else begin
        case (state)
            RESET: begin
                // reset everything in reset state
                grant = 0;
                s_counter = 0;
            end
            GRANT_PROCESS: begin
                // update mask
                // s_mask = nextGrant & (~nextGrant + 1);

                // granting logic
                grant = request & nextGrant & (~nextGrant + 1);

                // Initialize counter to 2 to account for the 3 clock cycle latency:
                // 1 cycle for request registration
                // 1 cycle for grant generation 
                // 1 cycle for weight loading
                // This ensures proper timing when weight is >= 2
                s_counter = 2;
            end
            GETWEIGHT: begin
                s_weight = weight;
            end
            COUNT: begin
                // count up until weight is reached account for clock cycle
                s_counter = s_counter + 1;

                // no change to grant
                grant = grant;
            end

            // if statemachine never goes out of wack
            default: begin
                grant = grant;
                s_counter = s_counter;
            end
        endcase
    end
end

`ifdef FORMAL
    reg f_valid_clock = 1'b0;
    always @(posedge clk) begin
        f_valid_clock = 1'b1;
    end

    reg f_valid_clock_2 = 1'b0;
    always @(posedge clk) begin
        if (f_valid_clock) begin
            f_valid_clock2 = 1'b1;
        end
    end

    always @(posedge clk) begin
        if (f_valid_clock_2) begin
            if (!reset) begin
                if (grant[1] == 1'b1) begin
                    assert($past(request[1]) == 1'b1);
                end
            end
        end
    end

    // Property 1: grant is always 0 if there is no request
    always @(posedge clk) begin
        if (f_valid_clock_2) begin
            // Check if request is all zeros using reduction OR operator
            // |request performs OR operation on all bits of request
            // If request is 0000, |request will be 0
            if (|grant == 0) begin
                assert(|$past(request) == 0);
            end
        end
    end

    always @(posedge clk) begin
        if (f_valid_clock_2) begin
            if (!reset && (request != 'b0) && (weight != 'b0)) begin
                assert ($onehot(grant));
            end
        end
    end
    
    // Property 2: grant is onehot 
    always @(posedge clk) begin
        if (f_valid_clock_2) begin
            assert ($onehot(grant));
        end
    end

    // Property 3: grant delay by weight
    reg [3:0] past_grant;
    reg [3:0] delay_count;
    
    always @(posedge clk) begin
        if (!reset) begin
            if (f_valid_clock_2) begin
                if (delay_count == s_weight) begin
                    assert(grant == past_grant);
                    past_grant <= grant;
                    delay_count <= 0;
                end else begin
                    delay_count <= delay_count + 1;
                end
            end
        end
    end

    // s_counter should be less than weight
    always @(posedge clk) begin
        if (f_valid_clock_2) begin
            assert(s_counter < s_weight);
        end
    end
    
    
    // property 4: s_counter should be greater than 0
    always @(posedge clk) begin
        if (f_valid_clock_2) begin
            assert(s_counter > 0);
        end
    end

    always @(posedge clk) begin
        assume(weight == 8);
    end

    initial begin
        s_counter = 0;
        state = GRANT_PROCESS;
        weight = 8;
    end

    // property 5: state change from GETWEIGHT to COUNT, s_counter should be 2
    always @(posedge clk) begin
        if (state == GRANT_PROCESS) begin
            assert(s_counter < weight);
        end
    end

    // property 6: 
`endif 

endmodule
